module LCD1602 #(
  parameter BITS = 8,
  parameter CLK_FREQ = 1
)(
  input                 i_lcd_clk,
  input      [BITS-1:0] i_o,
  output reg            o_rdy, //@md = 1'b0,
  output reg            o_en,
  output reg [8:0]      o_lcd
);

// STATE order: 0 -> 1 -> [ 2< -> 3 -> 1 >- ]
localparam [1:0] STATE_INIT     = 0;
localparam [1:0] STATE_DISPLAY  = 1;
localparam [1:0] STATE_WAIT     = 2;
localparam [1:0] STATE_PREP_LCD = 3;

reg [BITS-1:0] xfer1, i_o2;  // CDC transfer i_o
reg [BITS-1:0] obuf;    // buffer new incoming value to check if changed
reg [21:0] tick;        // 1 = 62.5ns, 1600000 = 100ms, 2048000 = 128ms
reg [1:0] state = 0;    // 0=init 1=display 2=wait 3=preplcd
reg [8:0] lcdbuf[0:79]; // one line lcd data top print
reg [23:0] val;         // copy of new value
reg [6:0]  dig;         // bcd digit of new val
reg [2:0]  digs;        // number of bcd digits of new val
reg [6:0]  i;

always @(negedge i_lcd_clk) begin
  {i_o2, xfer1} <= {xfer1, i_o};
  //o_rdy <= (state > STATE_INIT) ? 1'b1 : 1'b0;
  //o_rdy <= (obuf == i_o2 && state == STATE_WAIT) ? 1'b1 : 1'b0;
  case (state)
    STATE_INIT: begin  // boot: initialize LCD 0000000000000000000000000000000000000000000000000000000000000000
      case (tick) // 48/4 = 12 MHz
        2000000: o_lcd  <= 9'b0_00110000;       // 125ms: Function Set1: 1 lines, 5x7 matrix, 8-bit
        2000001: o_en <= 1'b1;
        2000005: o_en <= 1'b0;
        2080000: o_lcd  <= 9'b0_00110000;       // 5ms: Function Set2: 1 lines, 5x7 matrix, 8-bit
        2080001: o_en <= 1'b1;
        2080005: o_en <= 1'b0;
        2084000: o_lcd  <= 9'b0_00110000;       // 200us: Function Set3: 1 lines, 5x7 matrix, 8-bit
        2084001: o_en <= 1'b1;
        2084005: o_en <= 1'b0;
        2088000: o_lcd  <= 9'b0_00111000;       // 200us: Function Set4: 2 lines, 5x7 matrix, 8-bit
        2088001: o_en <= 1'b1;
        2088005: o_en <= 1'b0;
        2089000: o_lcd  <= 9'b0_00001000;       // +62.5us: Display off
        2089001: o_en <= 1'b1;
        2089005: o_en <= 1'b0;
        2090000: o_lcd  <= 9'b0_00000001;       // +62.5us: Clear display
        2090001: o_en <= 1'b1;
        2090005: o_en <= 1'b0;
        2170000: o_lcd  <= 9'b0_00000111;       // +5ms: Entry mode: Display shift right
        2170001: o_en <= 1'b1;
        2170005: o_en <= 1'b0;
        2171000: o_lcd  <= 9'b0_00001100;       // +62.5us: Display on no cursor
        2171001: o_en <= 1'b1;
        2171005: o_en <= 1'b0;
        2200000: begin                          // +1.64ms: Prepare 1st line
          // 0123456789012345
          // 16bit 16000000Hz
          if (BITS > 9) begin
            lcdbuf[0]  <= {1'b1,8'd49}; // 1
            lcdbuf[1]  <= {1'b1,8'd48} + BITS - 10; // 6
          end else begin
            lcdbuf[0]  <= {1'b1,8'd32}; // spc
            lcdbuf[1]  <= {1'b1,8'd48} + BITS; // 8
          end
          lcdbuf[2]  <= {1'b1,8'd98};   // b
          lcdbuf[3]  <= {1'b1,8'd105};  // i
          lcdbuf[4]  <= {1'b1,8'd116};  // t
          lcdbuf[5]  <= {1'b1,8'd32};   // spc
          val = CLK_FREQ;
          /* Binary to BCD (too big)
          for (i=23; i>=0; i=i-1) begin
        		for (dig=6; dig<=13; dig=dig+1) begin
        			if (lcdbuf[dig] >= 5) lcdbuf[dig] = lcdbuf[dig] +3;
        			lcdbuf[dig] = lcdbuf[dig] << 1;
        			if (dig<13)
        				lcdbuf[dig][0] = lcdbuf[dig+1][3];
        			else
        				lcdbuf[13][0] = val[i];
        		end
        	end
        	for (dig=6; dig<=15; dig=dig+1)
        		lcdbuf[dig] = lcdbuf[dig] + {1'b1,8'd48};
          */
          if     (val >= 24'd10000000) begin lcdbuf[6]  <= {1'b1,8'd49}; val = val - 24'd10000000; end
          else                               lcdbuf[6]  <= {1'b1,8'd32}; // spc
          if      (val >= 24'd9000000) begin lcdbuf[7]  <= {1'b1,8'd57}; val = val - 24'd9000000; end
          else if (val >= 24'd8000000) begin lcdbuf[7]  <= {1'b1,8'd56}; val = val - 24'd8000000; end
          else if (val >= 24'd7000000) begin lcdbuf[7]  <= {1'b1,8'd55}; val = val - 24'd7000000; end
          else if (val >= 24'd6000000) begin lcdbuf[7]  <= {1'b1,8'd54}; val = val - 24'd6000000; end
          else if (val >= 24'd5000000) begin lcdbuf[7]  <= {1'b1,8'd53}; val = val - 24'd5000000; end
          else if (val >= 24'd4000000) begin lcdbuf[7]  <= {1'b1,8'd52}; val = val - 24'd4000000; end
          else if (val >= 24'd3000000) begin lcdbuf[7]  <= {1'b1,8'd51}; val = val - 24'd3000000; end
          else if (val >= 24'd2000000) begin lcdbuf[7]  <= {1'b1,8'd50}; val = val - 24'd2000000; end
          else if (val >= 24'd1000000) begin lcdbuf[7]  <= {1'b1,8'd49}; val = val - 24'd1000000; end
          else if (val != CLK_FREQ)          lcdbuf[7]  <= {1'b1,8'd48};
          else                               lcdbuf[7]  <= {1'b1,8'd32};
          if      (val >= 24'd900000)  begin lcdbuf[8]  <= {1'b1,8'd57}; val = val - 24'd900000;  end
          else if (val >= 24'd800000)  begin lcdbuf[8]  <= {1'b1,8'd56}; val = val - 24'd800000;  end
          else if (val >= 24'd700000)  begin lcdbuf[8]  <= {1'b1,8'd55}; val = val - 24'd700000;  end
          else if (val >= 24'd600000)  begin lcdbuf[8]  <= {1'b1,8'd54}; val = val - 24'd600000;  end
          else if (val >= 24'd500000)  begin lcdbuf[8]  <= {1'b1,8'd53}; val = val - 24'd500000;  end
          else if (val >= 24'd400000)  begin lcdbuf[8]  <= {1'b1,8'd52}; val = val - 24'd400000;  end
          else if (val >= 24'd300000)  begin lcdbuf[8]  <= {1'b1,8'd51}; val = val - 24'd300000;  end
          else if (val >= 24'd200000)  begin lcdbuf[8]  <= {1'b1,8'd50}; val = val - 24'd200000;  end
          else if (val >= 24'd100000)  begin lcdbuf[8]  <= {1'b1,8'd49}; val = val - 24'd100000;  end
          else if (val != CLK_FREQ)          lcdbuf[8]  <= {1'b1,8'd48};
          else                               lcdbuf[8]  <= {1'b1,8'd32};
          if      (val >= 24'd90000)   begin lcdbuf[9]  <= {1'b1,8'd57}; val = val - 24'd90000;   end
          else if (val >= 24'd80000)   begin lcdbuf[9]  <= {1'b1,8'd56}; val = val - 24'd80000;   end
          else if (val >= 24'd70000)   begin lcdbuf[9]  <= {1'b1,8'd55}; val = val - 24'd70000;   end
          else if (val >= 24'd60000)   begin lcdbuf[9]  <= {1'b1,8'd54}; val = val - 24'd60000;   end
          else if (val >= 24'd50000)   begin lcdbuf[9]  <= {1'b1,8'd53}; val = val - 24'd50000;   end
          else if (val >= 24'd40000)   begin lcdbuf[9]  <= {1'b1,8'd52}; val = val - 24'd40000;   end
          else if (val >= 24'd30000)   begin lcdbuf[9]  <= {1'b1,8'd51}; val = val - 24'd30000;   end
          else if (val >= 24'd20000)   begin lcdbuf[9]  <= {1'b1,8'd50}; val = val - 24'd20000;   end
          else if (val >= 24'd10000)   begin lcdbuf[9]  <= {1'b1,8'd49}; val = val - 24'd10000;   end
          else if (val != CLK_FREQ)          lcdbuf[9]  <= {1'b1,8'd48};
          else                               lcdbuf[9]  <= {1'b1,8'd32};
          if      (val >= 24'd9000)    begin lcdbuf[10] <= {1'b1,8'd57}; val = val - 24'd9000;    end
          else if (val >= 24'd8000)    begin lcdbuf[10] <= {1'b1,8'd56}; val = val - 24'd8000;    end
          else if (val >= 24'd7000)    begin lcdbuf[10] <= {1'b1,8'd55}; val = val - 24'd7000;    end
          else if (val >= 24'd6000)    begin lcdbuf[10] <= {1'b1,8'd54}; val = val - 24'd6000;    end
          else if (val >= 24'd5000)    begin lcdbuf[10] <= {1'b1,8'd53}; val = val - 24'd5000;    end
          else if (val >= 24'd4000)    begin lcdbuf[10] <= {1'b1,8'd52}; val = val - 24'd4000;    end
          else if (val >= 24'd3000)    begin lcdbuf[10] <= {1'b1,8'd51}; val = val - 24'd3000;    end
          else if (val >= 24'd2000)    begin lcdbuf[10] <= {1'b1,8'd50}; val = val - 24'd2000;    end
          else if (val >= 24'd1000)    begin lcdbuf[10] <= {1'b1,8'd49}; val = val - 24'd1000;    end
          else if (val != CLK_FREQ)          lcdbuf[10] <= {1'b1,8'd48};
          else                               lcdbuf[10] <= {1'b1,8'd32};
          if      (val >= 24'd900)     begin lcdbuf[11] <= {1'b1,8'd57}; val = val - 24'd900;     end
          else if (val >= 24'd800)     begin lcdbuf[11] <= {1'b1,8'd56}; val = val - 24'd800;     end
          else if (val >= 24'd700)     begin lcdbuf[11] <= {1'b1,8'd55}; val = val - 24'd700;     end
          else if (val >= 24'd600)     begin lcdbuf[11] <= {1'b1,8'd54}; val = val - 24'd600;     end
          else if (val >= 24'd500)     begin lcdbuf[11] <= {1'b1,8'd53}; val = val - 24'd500;     end
          else if (val >= 24'd400)     begin lcdbuf[11] <= {1'b1,8'd52}; val = val - 24'd400;     end
          else if (val >= 24'd300)     begin lcdbuf[11] <= {1'b1,8'd51}; val = val - 24'd300;     end
          else if (val >= 24'd200)     begin lcdbuf[11] <= {1'b1,8'd50}; val = val - 24'd200;     end
          else if (val >= 24'd100)     begin lcdbuf[11] <= {1'b1,8'd49}; val = val - 24'd100;     end
          else if (val != CLK_FREQ)          lcdbuf[11] <= {1'b1,8'd48};
          else                               lcdbuf[11] <= {1'b1,8'd32};
          if      (val >= 24'd90)      begin lcdbuf[12] <= {1'b1,8'd57}; val = val - 24'd90;      end
          else if (val >= 24'd80)      begin lcdbuf[12] <= {1'b1,8'd56}; val = val - 24'd80;      end
          else if (val >= 24'd70)      begin lcdbuf[12] <= {1'b1,8'd55}; val = val - 24'd70;      end
          else if (val >= 24'd60)      begin lcdbuf[12] <= {1'b1,8'd54}; val = val - 24'd60;      end
          else if (val >= 24'd50)      begin lcdbuf[12] <= {1'b1,8'd53}; val = val - 24'd50;      end
          else if (val >= 24'd40)      begin lcdbuf[12] <= {1'b1,8'd52}; val = val - 24'd40;      end
          else if (val >= 24'd30)      begin lcdbuf[12] <= {1'b1,8'd51}; val = val - 24'd30;      end
          else if (val >= 24'd20)      begin lcdbuf[12] <= {1'b1,8'd50}; val = val - 24'd20;      end
          else if (val >= 24'd10)      begin lcdbuf[12] <= {1'b1,8'd49}; val = val - 24'd10;      end
          else if (val != CLK_FREQ)          lcdbuf[12] <= {1'b1,8'd48};
          else                               lcdbuf[12] <= {1'b1,8'd32};
          lcdbuf[13] <= {1'b1,8'd48} + val;
          lcdbuf[14] <= {1'b1,8'd72};   // H
          lcdbuf[15] <= {1'b1,8'd122};  // z
          //lcdbuf[16:79] <= {64{1'b1,8'd32}}; // spc
          for (i = 7'd16; i <= 7'd79; i = i + 7'd1)
            lcdbuf[i] <= {1'b1,8'd32}; // spc
          // Goto display state
          dig <= 7'd0;
          state <= STATE_DISPLAY;
        end
      endcase
      if (tick < 2200000)
        tick <= tick + 22'd1;
      else
        tick <= 22'd0;
    end
    STATE_DISPLAY: begin // update display 11111111111111111111111111111111111111111111111111111111111111111111
      case (tick)
        800: o_lcd <= lcdbuf[dig]; // (1700@13MHz) 800 375us = ticks between characters
        804: o_en <= 1'b1;         // (1710@13MHz) 804
        840: begin                 // (1800@13MHz) 840
          o_en <= 1'b0;
          if (dig < 7'd79)
            dig <= dig + 7'd1;
          else begin
            state <= STATE_WAIT;
            obuf <= i_o2;
          end
        end
      endcase
      if (tick < 22'd840) // (1800@13MHz) 840
        tick <= tick + 22'd1;
      else
        tick <= 22'd0;
    end
    STATE_WAIT: begin // waiting for new char in O-register 222222222222222222222222222222222222222222222222222
      if (tick < 22'd160000) // (275000@13MHz)  160000 10ms = ticks between display updates
        tick <= tick + 22'd1;
      else if (obuf != i_o2) begin
        o_rdy <= 1'b0;
        if      (i_o2 >= 10000) digs = 3'd6;
        else if (i_o2 >= 1000)  digs = 3'd5;
        else if (i_o2 >= 100)   digs = 3'd4;
        else if (i_o2 >= 10)    digs = 3'd3;
        else                    digs = 3'd2;
        for (i = 7'd40; i <= 7'd53; i = i + 7'd1)
          lcdbuf[i] <= lcdbuf[i+digs];
        state <= STATE_PREP_LCD;
      end else
        o_rdy <= 1'b1;
    end
    STATE_PREP_LCD: begin // prep chars in lcd-buffer 333333333333333333333333333333333333333333333333333333333
      val = i_o2;
      if      (val >= 24'd60000)   begin lcdbuf[51] <= {1'b1,8'd54}; val = val - 24'd60000;   end
      else if (val >= 24'd50000)   begin lcdbuf[51] <= {1'b1,8'd53}; val = val - 24'd50000;   end
      else if (val >= 24'd40000)   begin lcdbuf[51] <= {1'b1,8'd52}; val = val - 24'd40000;   end
      else if (val >= 24'd30000)   begin lcdbuf[51] <= {1'b1,8'd51}; val = val - 24'd30000;   end
      else if (val >= 24'd20000)   begin lcdbuf[51] <= {1'b1,8'd50}; val = val - 24'd20000;   end
      else if (val >= 24'd10000)   begin lcdbuf[51] <= {1'b1,8'd49}; val = val - 24'd10000;   end
      if      (val >= 24'd9000)    begin lcdbuf[52] <= {1'b1,8'd57}; val = val - 24'd9000;    end
      else if (val >= 24'd8000)    begin lcdbuf[52] <= {1'b1,8'd56}; val = val - 24'd8000;    end
      else if (val >= 24'd7000)    begin lcdbuf[52] <= {1'b1,8'd55}; val = val - 24'd7000;    end
      else if (val >= 24'd6000)    begin lcdbuf[52] <= {1'b1,8'd54}; val = val - 24'd6000;    end
      else if (val >= 24'd5000)    begin lcdbuf[52] <= {1'b1,8'd53}; val = val - 24'd5000;    end
      else if (val >= 24'd4000)    begin lcdbuf[52] <= {1'b1,8'd52}; val = val - 24'd4000;    end
      else if (val >= 24'd3000)    begin lcdbuf[52] <= {1'b1,8'd51}; val = val - 24'd3000;    end
      else if (val >= 24'd2000)    begin lcdbuf[52] <= {1'b1,8'd50}; val = val - 24'd2000;    end
      else if (val >= 24'd1000)    begin lcdbuf[52] <= {1'b1,8'd49}; val = val - 24'd1000;    end
      else if (val != i_o2)          lcdbuf[52] <= {1'b1,8'd48};
      if      (val >= 24'd900)     begin lcdbuf[53] <= {1'b1,8'd57}; val = val - 24'd900;     end
      else if (val >= 24'd800)     begin lcdbuf[53] <= {1'b1,8'd56}; val = val - 24'd800;     end
      else if (val >= 24'd700)     begin lcdbuf[53] <= {1'b1,8'd55}; val = val - 24'd700;     end
      else if (val >= 24'd600)     begin lcdbuf[53] <= {1'b1,8'd54}; val = val - 24'd600;     end
      else if (val >= 24'd500)     begin lcdbuf[53] <= {1'b1,8'd53}; val = val - 24'd500;     end
      else if (val >= 24'd400)     begin lcdbuf[53] <= {1'b1,8'd52}; val = val - 24'd400;     end
      else if (val >= 24'd300)     begin lcdbuf[53] <= {1'b1,8'd51}; val = val - 24'd300;     end
      else if (val >= 24'd200)     begin lcdbuf[53] <= {1'b1,8'd50}; val = val - 24'd200;     end
      else if (val >= 24'd100)     begin lcdbuf[53] <= {1'b1,8'd49}; val = val - 24'd100;     end
      else if (val != i_o2)          lcdbuf[53] <= {1'b1,8'd48};
      if      (val >= 24'd90)      begin lcdbuf[54] <= {1'b1,8'd57}; val = val - 24'd90;      end
      else if (val >= 24'd80)      begin lcdbuf[54] <= {1'b1,8'd56}; val = val - 24'd80;      end
      else if (val >= 24'd70)      begin lcdbuf[54] <= {1'b1,8'd55}; val = val - 24'd70;      end
      else if (val >= 24'd60)      begin lcdbuf[54] <= {1'b1,8'd54}; val = val - 24'd60;      end
      else if (val >= 24'd50)      begin lcdbuf[54] <= {1'b1,8'd53}; val = val - 24'd50;      end
      else if (val >= 24'd40)      begin lcdbuf[54] <= {1'b1,8'd52}; val = val - 24'd40;      end
      else if (val >= 24'd30)      begin lcdbuf[54] <= {1'b1,8'd51}; val = val - 24'd30;      end
      else if (val >= 24'd20)      begin lcdbuf[54] <= {1'b1,8'd50}; val = val - 24'd20;      end
      else if (val >= 24'd10)      begin lcdbuf[54] <= {1'b1,8'd49}; val = val - 24'd10;      end
      else if (val != i_o2)          lcdbuf[54] <= {1'b1,8'd48};
                                     lcdbuf[55] <= {1'b1,8'd48} + val;
      dig <= 7'd0;
      tick <= 22'd0;
      state <= STATE_DISPLAY;
    end
  endcase
end
endmodule
